     /**
      * \brief List Implementation : Vector
      *
      * A list implementation with a memory block as backend. The memory for
      * will be dynamically allocated and deallocated. Therefore the list
      * can grow and shrink dynamically as requested.
      *
      * <br><br>
      *
      * A user created storage heap will be used for memory allocation /
      * deallocation.
      *
      * <br><br>
      *
      * This list implementation works with a head data structure.
      *
      * <br><br>
      *
      * The entries are stored in an "array" which consists of pointers.
      * The pointers of the array store the start address of the memory
      * of the values.
      *
      * <br><br>
      *
      * All values are internally null-terminated. So a value of x'00'
      * won't work as expected and should be avoided.
      *
      * <br><br>
      *
      * Access to the element is accomplished through accessing the vector
      * with an index (position). The index is 0-based. So the first element
      * has an index of 0 (zero).
      *
      * \author Mihael Schmidt
      * \date   23.06.2008
      */



      *------------------------------------------------------------------------
      *
      * Copyright (c) 2007-2009 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the VECTOR service program.
      *
      * VECTOR is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * VECTOR is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with VECTOR.  If not, see http://www.gnu.org/licenses/.
      *
      *------------------------------------------------------------------------



     HNOMAIN
     HBNDDIR('QC2LE')
     HCOPYRIGHT('Copyright (c) 2007-2009 Mihael Schmidt. All rights reserved.')
     HDEBUG(*YES)
     HOPTION(*SRCSTMT)


      *---------------------------------------------------------------
      * Prototypen
      *---------------------------------------------------------------
     D increment       PR
     D   vector                        *   const
     D   size                        10U 0 const options(*nopass)
      *
     D memcpy          PR                  extproc('memcpy')
     D   target                        *   value
     D   source                        *   value
     D   length                      10U 0 value
      *
     D memcmp          PR            10I 0 extproc('memcmp')
     D   buffer1                       *   value
     D   buffer2                       *   value
     D   length                      10U 0 value
      *
     D getToken        PR              *   extproc('strtok')
     D   string                        *   value options(*string)
     D   delim                         *   value options(*string)
      *
     D strlen...
     D                 PR            10U 0 extproc('strlen')
     D   string                        *   value
      *
     D sendEscapeMessage...
     D                 PR
     D   id                          10I 0 const
      *
     D getVectorEntry...
     D                 PR              *
     D   vector                        *   const
     D   pos                         10I 0 const
      *
     D isVectorImpl...
     D                 PR
     D   vector                        *   const
      *
     D moveElements...
     D                 PR
     D   vector                        *   const
     D   pos                         10U 0 const
     D   pDownBy                     10I 0 const options(*nopass)

      /copy VECTOR_H
      /copy CEEAPI_H


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
     D VECTOR_ID       C                   'VECTOR_IMPL.'
      *
      * Incrementation size of 0 (zero) doubles the size of the array.
      *
     D DEFAULT_INCREMENT_SIZE...
     D                 C                    0
     D DEFAULT_INIT_SIZE...
     D                 C                   10
      *
      * Message IDs
      *
     D MSG_NO_VECTOR_IMPL...
     D                 C                   1
     D MSG_POSITION_OUT_OF_BOUNDS...
     D                 C                   2
     D MSG_MAX_SIZE_REACHED...
     D                 C                   3
     D MSG_DEST_OUT_OF_BOUNDS...
     D                 C                   4
     D MSG_RANGE_OUT_OF_BOUNDS...
     D                 C                   5
     D MSG_INVALID_VALUE_TYPE...
     D                 C                   6


      *-------------------------------------------------------------------------
      * Variables
      *-------------------------------------------------------------------------
     D tmpl_header     DS                  qualified based(nullPointer)
     D   id                          20A
     D   heapId                      10I 0
     D   elementsAllocated...
     D                               10U 0
     D   elementCount                10U 0
     D   elementData                   *
     D   incrementSize...
     D                               10U 0
      *
     D tmpl_entry      DS                  qualified
     D   valuePtr                      *
      *
     D hexNull         S              1A   inz(x'00')


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------

     /**
      * \brief Create vector
      *
      * Creates a vector.
      *
      * <br><br>
      *
      * The initial size is 10. The default increment size
      * is 0 which means with each incrementation the vector will double its
      * size.
      *
      * \author Mihael Schmidt
      * \date 04.07.2008
      *
      * \param Initial vector size (default: 10)
      * \param Incrementation size (default: 0 - double)
      *
      * \return Pointer to vector
      */
     P vector_create...
     P                 B                   export
     D                 PI              *
     D   initSize                    10U 0 const options(*nopass)
     D   incSize                     10U 0 const options(*nopass)
      *
     D vector          S               *
     D header          DS                  likeds(tmpl_header) based(vector)
     D heapId          S             10I 0
      /free
       cee_createHeap(heapId : *omit : *omit : *omit : *omit);

       // allocate memory
       cee_getStorage(heapId : %size(tmpl_header) : vector : *omit);

       header.id = VECTOR_ID;
       header.heapId = heapId;
       header.elementCount = 0;
       header.elementData = *null;

       // init vector size if passed
       if (%parms() = 0);
         header.elementsAllocated = DEFAULT_INIT_SIZE;
         header.incrementSize = DEFAULT_INCREMENT_SIZE;
       elseif (%parms() = 1);
         header.incrementSize = DEFAULT_INCREMENT_SIZE;
         header.elementsAllocated = initSize;
       elseif (%parms() = 2);
         header.incrementSize = incSize;
         header.elementsAllocated = initSize;
       endif;

       // alloc memory for entries
       cee_getStorage(heapId :
                      header.elementsAllocated * %size(header.elementData) :
                      header.elementData :
                      *omit);

       return vector;
      /end-free
     P                 E


     /**
      * \brief Dispose vector
      *
      * Disposes the vector and all its elements. The pointer will be set
      * to *null.
      *
      * \author Mihael Schmidt
      * \date   04.07.2008
      *
      * \param Pointer to vector
      */
     P vector_dispose...
     P                 B                   export
     D                 PI
     D   vector                        *
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D i               S             10I 0
      /free
       if (vector <> *null);
         isVectorImpl(vector);
         cee_discardHeap(header.heapId : *omit);
         vector = *null;
       endif;
      /end-free
     P                 E


     /**
      * \brief Add element
      *
      * Adds an element to the vector by copying the content to dynamically
      * allocated memory. Values are stored null-terminated.
      *
      * <br><br>
      *
      * If a position is passed the caller must be certain that the
      * position is inside the bounds of the vector. If the position is outside
      * of the vector an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   04.07.2008
      *
      * \param Pointer to vector
      * \param Pointer to new entry
      * \param Length of new entry (in byte)
      * \param Position
      */
     P vector_add      B                   export
     D                 PI
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
     D   pPos                        10U 0 const options(*nopass)
      *
     D pos             S             10U 0
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataPtr         S               *   based(ptr)
      /free
       isVectorImpl(vector);

       // determine the position
       if (%parms() = 3);
         pos = header.elementCount;
       elseif (%parms() = 4);
         // if a position was passed, check if it is inside the vector
         if (pPos >= header.elementCount);
           sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         else;
           pos = pPos;
           moveElements(vector : pos);
         endif;
       endif;

       // make some space if necessary
       if (header.elementCount = header.elementsAllocated);
         increment(vector);
       endif;

       // get free entry
       ptr = getVectorEntry(vector : pos);

       // alloc memory for this entry
       cee_getStorage(header.heapId :
                      valueLength + 1 : // +1 for the null byte
                      dataPtr :
                      *omit);

       // copy value to the entry
       memcpy(dataPtr : value : valueLength);

       // set null to the last byte
       memcpy(dataPtr + valueLength : %addr(hexNull) : 1);

       // update header
       header.elementCount += 1;
      /end-free
     P                 E


     /**
      * \brief Append element to the vector
      *
      * Adds an element to the end of the vector by copying the content to
      * dynamically allocated memory. Values are stored null-terminated.
      *
      * \author Mihael Schmidt
      * \date   27.07.2008
      *
      * \param Pointer to vector
      * \param Pointer to new entry
      * \param Length of new entry (in byte)
      */
     P vector_addLast...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
      /free
       vector_add(vector : value : valueLength);
      /end-free
     P                 E


     /**
      * \brief Prepend element to the vector
      *
      * Adds an element to the beginning of the vector by copying the content to
      * dynamically allocated memory. Values are stored null-terminated. If the
      * the vector is not empty all other elements will be pushed down by one
      * position.
      *
      * \author Mihael Schmidt
      * \date   30.07.2008
      *
      * \param Pointer to vector
      * \param Pointer to new entry
      * \param Length of new entry (in byte)
      */
     P vector_addFirst...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
      /free
       vector_add(vector : value : valueLength : 0);
      /end-free
     P                 E

     /**
      * \brief Get vector size
      *
      * Returns the number of elements currently in the vector.
      *
      * \author Mihael Schmidt
      * \date   24.07.2008
      *
      * \param Pointer to vector
      *
      * \return Number of elements in the vector
      */
     P vector_getSize...
     P                 B                   export
     D                 PI            10U 0
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D size            S             10U 0
      /free
       isVectorImpl(vector);

       return header.elementCount;
      /end-free
     P                 E


     /**
      * \brief Get vector capacity
      *
      * Returns the number of elements which can be stored in the current
      * vector.
      *
      * \author Mihael Schmidt
      * \date   24.07.2008
      *
      * \param Pointer to vector
      *
      * \return Number of elements able to store in the vector
      */
     P vector_getCapacity...
     P                 B                   export
     D                 PI            10U 0
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D size            S             10U 0
      /free
       isVectorImpl(vector);

       return header.elementsAllocated;
      /end-free
     P                 E



     /**
      * \brief Get element
      *
      * Returns a pointer to the elment at the given position. The element is
      * null-terminated. Changes to the element through the returned pointer is
      * not recommended. Use the appropriate procedures instead.
      *
      * <br><br>
      *
      * If the requested element position is not in the vector then an escape
      * message will be sent.
      *
      * \author Mihael Schmidt
      * \date   05.07.2008
      *
      * \param Pointer to vector
      * \param Position
      *
      * \return Pointer to the null-terminated element
      *         or *null if vector is empty
      */
     P vector_get      B                   export
     D                 PI              *
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataPtr         S               *   based(ptr)
      /free
       isVectorImpl(vector);

       if (header.elementCount = 0);
         return *null;
       elseif (index >= header.elementCount);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
       endif;

       ptr = getVectorEntry(vector : index);

       return dataPtr;
      /end-free
     P                 E


     /**
      * \brief Get first element
      *
      * Returns a pointer to the first elment in the vector. The element is
      * null-terminated. Changes to the element through the returned pointer is
      * not recommended. Use the appropriate procedures instead.
      *
      * \author Mihael Schmidt
      * \date   03.08.2008
      *
      * \param Pointer to vector
      *
      * \return Pointer to the null-terminated element or *null if the vector is empty
      */
     P vector_getFirst...
     P                 B                   export
     D                 PI              *
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
      /free
       return vector_get(vector : 0);
      /end-free
     P                 E


     /**
      * \brief Get last element
      *
      * Returns a pointer to the last elment in the vector. The element is
      * null-terminated. Changes to the element through the returned pointer is
      * not recommended. Use the appropriate procedures instead.
      *
      * \author Mihael Schmidt
      * \date   03.08.2008
      *
      * \param Pointer to vector
      *
      * \return Pointer to the null-terminated element
      *         or *null if the vector is empty
      */
     P vector_getLast...
     P                 B                   export
     D                 PI              *
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
      /free
       return vector_get(vector : header.elementCount - 1);
      /end-free
     P                 E


     /**
      * \brief Check if vector is empty
      *
      * Checks if the vector is empty.
      *
      * \author Mihael Schmidt
      * \date   27.07.2008
      *
      * \param Pointer to the vector
      *
      * \return *on = vector is empty <br>
      *         *off = vector is not empty
      */
     P vector_isEmpty...
     P                 B                   export
     D                 PI              N
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
      /free
       isVectorImpl(vector);
       return (header.elementCount = 0);
      /end-free
     P                 E


     /**
      * \brief Clear vector
      *
      * Deletes all entries. The capacity of the vector remains the same.
      *
      * \author Mihael Schmidt
      * \date   27.07.2008
      *
      * \param Pointer to the vector
      */
     P vector_clear    B                   export
     D                 PI
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D i               S             10I 0
     D ptr             S               *
     D dataPtr         S               *   based(ptr)
      /free
       isVectorImpl(vector);

       for i = 0 to header.elementCount - 1;
         ptr = header.elementData + (i * %size(tmpl_entry));
         cee_freeStorage(dataPtr : *omit);
       endfor;

       header.elementCount = 0;
      /end-free
     P                 E


     /**
      * \brief Remove an element
      *
      * Removes an element from the vector. If the given position is outside of
      * the bounds of the vector an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   01.08.2008
      *
      * \param Pointer to the vector
      * \param Element index to be removed
      */
     P vector_remove...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataPtr         S               *   based(ptr)
      /free
       isVectorImpl(vector);

       if (index >= header.elementCount);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
       endif;

       ptr = getVectorEntry(vector : index);
       cee_freeStorage(dataPtr : *omit); // release allocated memory
       dataPtr = *null;

       // check if it is the last element
       if (index <> header.elementCount -1); // not the last => move other elements
         moveElements(vector : index + 1 : -1);
       endif;

       header.elementCount -= 1;
      /end-free
     P                 E


     /**
      * \brief Remove the first element
      *
      * Removes the first element from the vector.
      *
      * \author Mihael Schmidt
      * \date   01.08.2008
      *
      * \param Pointer to the vector
      */
     P vector_removeFirst...
     P                 B                   export
     D                 PI
     D   vector                        *   const
      /free
       vector_remove(vector : 0);
      /end-free
     P                 E


     /**
      * \brief Remove the last element
      *
      * Removes the last element from the vector.
      *
      * \author Mihael Schmidt
      * \date   01.08.2008
      *
      * \param Pointer to the vector
      */
     P vector_removeLast...
     P                 B                   export
     D                 PI
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
      /free
       if (header.elementCount > 0);
         vector_remove(vector : vector_getSize(vector) - 1);
       endif;
      /end-free
     P                 E


     /**
      * \brief Remove a range of elements
      *
      * Removes a range of elements from the vector. The range must be inside
      * the bounds of the vector. If the range is outside the vector an escape
      * message will be sent.
      *
      * \author Mihael Schmidt
      * \date   01.08.2008
      *
      * \param Pointer to the vector
      * \param Range starting index
      * \param Number of elements to remove
      */
     P vector_removeRange...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   index                       10U 0 const
     D   count                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D i               S             10I 0
      /free
       isVectorImpl(vector);

       if (index + count -1 >= header.elementCount);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
       endif;

       for i = index to index + count - 1;
         // always index for position because the other elements have moved up by one
         vector_remove(vector : index);
       endfor;

      /end-free
     P                 E


     /**
      * \brief Contains element
      *
      * Checks if the vector contains the passed data.
      * The check will be done byte by byte, so trailing spaces also count.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      * \param Pointer to data
      * \param Data length
      *
      * \return *on if the vector contains the data, *off otherwise
      */
     P vector_contains...
     P                 B                   export
     D                 PI              N
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
     D i               S             10I 0
     D retVal          S               N   inz(*off)
      /free
       isVectorImpl(vector);

       for i = 0 to header.elementCount - 1;
         ptr = header.elementData + (i * %size(tmpl_entry));
         if (strlen(data) = valueLength);

           if (memcmp(data : value : valueLength) = 0);
             retVal = *on;
             leave;
           endif;

         endif;
       endfor;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get index of element
      *
      * Returns the index of the passed element.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      * \param Pointer to data
      * \param Data length
      *
      * \return index of the element or -1 if the element is not in the vector
      */
     P vector_indexOf...
     P                 B                   export
     D                 PI            10I 0
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
     D i               S             10I 0
     D retVal          S             10I 0 inz(-1)
      /free
       isVectorImpl(vector);

       for i = 0 to header.elementCount - 1;
         ptr = header.elementData + (i * %size(tmpl_entry));
         if (strlen(data) = valueLength);

           if (memcmp(data : value : valueLength) = 0);
             retVal = i;
             leave;
           endif;

         endif;
       endfor;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get last index of element
      *
      * Returns the last index of the passed element.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      * \param Pointer to data
      * \param Data length
      *
      * \return last index of the element
      *         or -1 if the element is not in the vector
      */
     P vector_lastIndexOf...
     P                 B                   export
     D                 PI            10I 0
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
     D i               S             10I 0
     D retVal          S             10I 0 inz(-1)
      /free
       isVectorImpl(vector);

       for i = header.elementCount - 1 downto 0;
         ptr = header.elementData + (i * %size(tmpl_entry));
         if (strlen(data) = valueLength);

           if (memcmp(data : value : valueLength) = 0);
             retVal = i;
             leave;
           endif;

         endif;
       endfor;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Frequency of element
      *
      * Returns the number of times the passed element is in the vector.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      * \param Pointer to data
      * \param Data length
      *
      * \return frequency of the passed data in the vector
      */
     P vector_frequency...
     P                 B                   export
     D                 PI            10U 0
     D   vector                        *   const
     D   value                         *   const
     D   valueLength                 10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
     D i               S             10I 0
     D retVal          S             10U 0
      /free
       isVectorImpl(vector);

       for i = 0 to header.elementCount - 1;
         ptr = header.elementData + (i * %size(tmpl_entry));
         if (strlen(data) = valueLength);

           if (memcmp(data : value : valueLength) = 0);
             retVal += 1;
           endif;

         endif;
       endfor;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Replace element
      *
      * Replaces the given element with the new data.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      * \param Index to data which should be replaced
      * \param Pointer to the new data
      * \param Length of the new data
      */
     P vector_replace...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   index                       10U 0 const
     D   value                         *   const
     D   valueLength                 10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
      /free
       isVectorImpl(vector);

       // if a position was passed, check if it is inside the array
       if (index >= header.elementCount);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
       endif;

       // get entry
       ptr = getVectorEntry(vector : index);

       // alloc memory for this entry
       cee_reallocateStorage(
                      data :
                      valueLength + 1 : // +1 for the null byte
                      *omit);

       // copy value to the entry
       memcpy(data : value : valueLength);

       // set null to the last byte
       memcpy(data + valueLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Create a sublist
      *
      * Returns a sublist of this vector.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      *
      * \return Pointer to the new vector (sublist)
      */
     P vector_sublist...
     P                 B                   export
     D                 PI              *
     D   vector                        *   const
     D   startIndex                  10U 0 const
     D   length                      10U 0 const options(*nopass)
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
      *
     D i               S             10U 0
     D endIndex        S             10U 0
     D sublist         S               *
      /free
       isVectorImpl(vector);

       sublist = vector_create();

       endIndex = startIndex + length;

       for i = startIndex to endIndex;
         ptr = getVectorEntry(vector : i);
         vector_add(sublist : data : strlen(data));
       endfor;

       return sublist;
      /end-free
     P                 E


     /**
      * \brief Create a copy of the vector
      *
      * Returns a copy of the vector.
      *
      * \author Mihael Schmidt
      * \date   26.11.2008
      *
      * \param Pointer to the vector
      *
      * \return Pointer to the new vector
      */
     P vector_copy...
     P                 B                   export
     D                 PI              *
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
      *
     D i               S             10U 0
     D copyList        S               *
      /free
       isVectorImpl(vector);

       copyList = vector_create();

       for i = 0 to header.elementCount -1;
         ptr = getVectorEntry(vector : i);
         vector_add(copyList : data : strlen(data));
       endfor;

       return copyList;
      /end-free
     P                 E


     /**
      * \brief Swap vector items
      *
      * \author Mihael Schmidt
      * \date   11.12.2008
      *
      * \param Pointer to the vector
      * \param Item to swap
      * \param Item to swap
      */
     P vector_swap...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   itemPos1                    10U 0 const
     D   itemPos2                    10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr1            S               *
     D data1           S               *   based(ptr1)
     D ptr2            S               *
     D data2           S               *   based(ptr2)
     D tmpPtr          S               *
      /free
       isVectorImpl(vector);

       // if a position was passed, check if it is inside the array
       if (itemPos1 >= header.elementCount or
           itemPos2 >= header.elementCount);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
       endif;

       ptr1 = getVectorEntry(vector : itemPos1);
       ptr2 = getVectorEntry(vector : itemPos2);

       tmpPtr = data1;
       data1 = data2;
       data2 = tmpPtr;
      /end-free
     P                 E


     /**
      * \brief Add all elements to the vector
      *
      * Adds all elements from the source vector to the destination vector.
      *
      * \param Pointer to the destination vector
      * \param Pointer to the source vector
      */
     P vector_addAll...
     P                 B                   export
     D                 PI
     D   destVector                    *   const
     D   sourceVector...
     D                                 *   const
      *
     D header          DS                  likeds(tmpl_header)
     D                                     based(sourceVector)
     D ptr             S               *
     D data            S               *   based(ptr)
      *
     D i               S             10U 0
      /free
       isVectorImpl(destVector);
       isVectorImpl(sourceVector);

       for i = 0 to header.elementCount -1;
         ptr = getVectorEntry(destVector : i);
         vector_add(sourceVector : data : strlen(data));
       endfor;
      /end-free
     P                 E


     /**
      * \brief Add integer value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds an integer to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addInteger...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                       10I 0 const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S             10I 0
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add short integer value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a short integer to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addShort...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                        5I 0 const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S              5I 0
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add long integer value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a long integer to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addLong...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                       20I 0 const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S             20I 0
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add packed decimal value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a packed decimal to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addDecimal...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                       15P 5 const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S             15P 5
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add float value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a float to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addFloat...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                        4F   const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S              4F
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add double value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a double to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addDouble...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                        8F   const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S              8F
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add date value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a date to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addDate...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                         D   const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S               D
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add boolean value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a boolean to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addBoolean...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   value                         N   const
     D   index                       10U 0 const options(*nopass)
      *
     D tmpValue        S               N
      /free
       tmpValue = value;

       if (%parms() = 2);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue));
       elseif (%parms() = 3);
         vector_add(vector : %addr(tmpValue) : %size(tmpValue) : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Add character value to the vector
      *
      * This procedure is a wrapper for the <em>add</em> procedure and
      * adds a character string to the vector.
      *
      * \param Pointer to the vector
      * \param Value
      * \param Position (default: append)
      */
     P vector_addString...
     P                 B                   export
     D                 PI                  opdesc
     D   vector                        *   const
     D   value                    65535A   const options(*varsize)
     D   index                       10U 0 const options(*nopass)
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0
      *
     D string          S          65535A   inz(*blank)
      /free
       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       string = %subst(value : 1 : length);

       if (%parms() = 2);
         vector_add(vector : %addr(string) : length);
       elseif (%parms() = 3);
         vector_add(vector : %addr(string) : length : index);
       endif;

      /end-free
     P                 E


     /**
      * \brief Get integer value from vector
      *
      * Returns the previously inserted integer value from the vector.
      * If the value cannot be interpreted as an integer an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getInteger...
     P                 B                   export
     D                 PI            10I 0
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S             10I 0 based(dataPtr)
     D retVal          S             10I 0
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get short integer value from vector
      *
      * Returns the previously inserted short integer value from the vector.
      * If the value cannot be interpreted as a short integer an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getShort...
     P                 B                   export
     D                 PI             5I 0
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S              5I 0 based(dataPtr)
     D retVal          S              5I 0
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get long integer value from vector
      *
      * Returns the previously inserted long integer value from the vector.
      * If the value cannot be interpreted as a long integer an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getLong...
     P                 B                   export
     D                 PI            20I 0
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S             20I 0 based(dataPtr)
     D retVal          S             20I 0
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get packed decimal value from vector
      *
      * Returns the previously inserted packed decimal value from the vector.
      * If the value cannot be interpreted as a packed decimal an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getDecimal...
     P                 B                   export
     D                 PI            15P 5
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S             15P 5 based(dataPtr)
     D retVal          S             15P 5
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get float value from vector
      *
      * Returns the previously inserted float value from the vector.
      * If the value cannot be interpreted as a float an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getFloat...
     P                 B                   export
     D                 PI             4F
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S              4F   based(dataPtr)
     D retVal          S              4F
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get dobule value from vector
      *
      * Returns the previously inserted double value from the vector.
      * If the value cannot be interpreted as a double an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getDouble...
     P                 B                   export
     D                 PI             8F
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S              8F   based(dataPtr)
     D retVal          S              8F
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get date value from vector
      *
      * Returns the previously inserted date value from the vector.
      * If the value cannot be interpreted as a date an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getDate...
     P                 B                   export
     D                 PI              D
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S               D   based(dataPtr)
     D retVal          S               D
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get boolean value from vector
      *
      * Returns the previously inserted boolean value from the vector.
      * If the value cannot be interpreted as a boolean an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getBoolean...
     P                 B                   export
     D                 PI              N
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D data            S               N   based(dataPtr)
     D retVal          S               N
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = data;
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Get character value from vector
      *
      * Returns the previously inserted character string value from the vector.
      * If the value cannot be interpreted as a char value an escape message
      * will be sent.
      *
      * \param Pointer to the vector
      * \param Position
      *
      * \return Value
      */
     P vector_getString...
     P                 B                   export
     D                 PI         65535A
     D   vector                        *   const
     D   index                       10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D dataptr         S               *   based(ptr)
     D retVal          S          65535A
      /free
       isVectorImpl(vector);

       // check if the vector is empty or the position is outside of the vector
       if (header.elementCount = 0);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       elseif (index < 0 or index > header.elementCount -1);
         sendEscapeMessage(MSG_POSITION_OUT_OF_BOUNDS);
         return *loval;
       endif;

       ptr = getVectorEntry(vector : index);

       monitor;
         // test if the temp variable is filled with the right data for the type
         // by moving the data from temp to another var (retVal in this case)
         retVal = %str(dataPtr);
         return retVal;
         on-error *all;
           sendEscapeMessage(MSG_INVALID_VALUE_TYPE);
           return *loval;
       endmon;

      /end-free
     P                 E


     /**
      * \brief Execute procedure for every vector entry
      *
      * The passed procedure will be executed for every entry
      * in the vector.
      *
      * <br><br>
      *
      * The user can pass data through a pointer to the procedure.
      * The pointer will not be touched by this procedure itself, so it
      * can be *null.
      *
      * <br><br>
      *
      * The value of list entry can be changed through the passed procedure.
      *
      * <br><br>
      *
      * The parameters for the passed procedure are:
      * <ul>
      *   <li>Pointer to the entry value (const)</li>
      *   <li>Pointer to the user data (const)</li>
      * </ul>
      *
      * \author Mihael Schmidt
      * \date   20.01.2009
      *
      * \param Pointer to the vector
      * \param Procedure pointer
      * \param Pointer to user data
      */
     P vector_foreach...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   procPtr                       *   const procptr
     D   userData                      *   const
      *
     D foreachProc     PR                  extproc(procPtr)
     D   valuePtr                      *   const
     D   userData                      *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr             S               *
     D data            S               *   based(ptr)
     D i               S             10I 0
      /free
       isVectorImpl(vector);

       for i = 0 to header.elementCount - 1;
         ptr = getVectorEntry(vector : i);
         foreachProc(data : userData);
       endfor;

      /end-free
     P                 E

     /**
      * \brief To character array
      *
      * Copies all entries of this vector to the passed array. Entries will be
      * truncated if they are too big for the array. If the array is not big
      * enough, the last entries will be silently dropped.
      *
      * \author Mihael Schmidt
      * \date   23.01.2009
      *
      * \param Pointer to the vector
      * \param Pointer to the array
      * \param Element count
      * \param Element length
      *
      */
     P vector_toCharArray...
     P                 B                   export
     D                 PI
     D   vector                        *   const
     D   arrayPtr                      *   const
     D   count                       10U 0 const
     D   length                      10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D entryPtr        S               *
     D dataPtr         S               *   based(entryPtr)
     D i               S             10I 0
     D arrayElemPtr    S               *
     D tmpLength       S             10I 0
      /free
       isVectorImpl(vector);

       for i = 0 to header.elementCount - 1;
         if (count > i);
           entryPtr = getVectorEntry(vector : i);
           arrayElemPtr = arrayPtr + (i * length);

           // determine how many bytes to copy (check length)
           if (strlen(dataPtr) < length);
             tmpLength = strlen(dataPtr);
           else;
             tmpLength = length;
           endif;

           // copy entry from vector to array
           memcpy(arrayElemPtr : dataPtr : tmpLength);
          endif;
       endfor;

      /end-free
     P                 E


     /**
      * \brief Reverse order of vector entries
      *
      * Reverses the order of the entries of the vector.
      *
      * \author Mihael Schmidt
      * \date   24.01.2009
      *
      * \param Pointer to the vector
      */
     P vector_reverse...
     P                 B                   export
     D                 PI
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D ptr1            S               *
     D data1           S               *   based(ptr1)
     D ptr2            S               *
     D data2           S               *   based(ptr2)
     D tmpPtr          S               *
     D i               S             10I 0
     D endValue        S             10I 0
      /free
       isVectorImpl(vector);

       if (header.elementCount <= 1);
         return;
       endif;

       endValue = header.elementCount / 2 - 1;

       for i = 0 to endValue;
         ptr1 = getVectorEntry(vector : i);
         ptr2 = getVectorEntry(vector : header.elementCount - i);

         tmpPtr = data1;
         data1 = data2;
         data2 = tmpPtr;
       endfor;

      /end-free
     P                 E


     /**
      * \brief Split character string
      *
      * The passed character string will be split into tokens by either
      * a passed or the default separator. All tokens will be added to
      * a new vector which will be returned.
      *
      * <br><br>
      *
      * Empty (but not blank) values will be dropped silently.
      *
      * \author Mihael Schmidt
      * \date   26.01.2009
      *
      * \param Character string (null-terminated)
      * \param Separator (default: ;)
      *
      * \return Pointer to the filled vector
      */
     P vector_split    B                   export
     D                 PI              *   opdesc
     D   pString                  65535A   const options(*varsize)
     D   pSeparator                   1A   const options(*nopass)
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0
      *
     D vector          S               *
     D token           S               *
     D separator       S              1A   inz(';')
     D string          S          65535A   inz(*blank)
      /free
       cee_getOpDescInfo(1 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       string = %subst(pString : 1 : length);

       if (%parms() = 2);
         separator = pSeparator;
       endif;

       vector = vector_create();

       token = getToken(string : separator);
       dow (token <> *null);
         vector_add(vector : token : strlen(token));
         token = getToken(*null : separator);
       enddo;

       return vector;
      /end-free
     P                 E


     /**
      * \brief Return character representation of vector
      *
      * Returns a string with the vector items separated either by
      * the passed or default separator. The items can be
      * enclosed by a passed character. The maximum character length
      * returned is 65535. Every character/item after that will be
      * dropped silently. Items will not be trimmed for this operation.
      *
      * <br><br>
      *
      * If the third parameter is passed, the third parameter will be
      * pre- and appended to the item. If the fourth parameter is also
      * passed the third parameter will be prepended to the item and the
      * fourth parameter will be appended to the item.
      *
      * \author Mihael Schmidt
      * \date   26.01.2009
      *
      * \param Pointer to the vector
      * \param separator (default: ,)
      * \param enclosing character (default: nothing)
      * \param enclosing character at the end of item (default: nothing)
      *
      * \return character representation of all vector items
      */
     P vector_toString...
     P                 B                   export
     D                 PI         65535A   varying
     D   vector                        *   const
     D   pSeparator                   1A   const options(*omit : *nopass)
     D   pEnclosing                 100A   const varying options(*nopass)
     D   pEnclosingEnd...
     D                              100A   const varying options(*nopass)
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D separator       S              1A   inz(',')
     D enclosingStart...
     D                 S            100A   varying
     D enclosingEnd...
     D                 S            100A   varying
     D entryPtr        S               *
     D valuePtr        S               *   based(entryPtr)
     D i               S             10I 0
     D retVal          S          65535A   varying
      /free
       isVectorImpl(vector);

       // check if separator is passed
       if (%parms() >= 2 and %addr(pSeparator) <> *null);
         separator = pSeparator;
       endif;

       // check if enclosing characters are passed
       if (%parms() >= 3);
         enclosingStart = pEnclosing;
         enclosingEnd = pEnclosing;
       endif;

       // check if we should use different chars for start and end of item
       if (%parms() = 4);
         enclosingEnd = pEnclosingEnd;
       endif;

       // process vector items
       for i = 0 to header.elementCount;
         entryPtr = getVectorEntry(vector : i);
         retVal += enclosingStart + %str(valuePtr) + enclosingEnd +
                   separator;
       endfor;

       // remove the last separator
       if (%len(retVal) > 0);
         %len(retVal) = %len(retVal) -1;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get entry pointer
      *
      * Returns the pointer of an entry.
      *
      * \author Mihael Schmidt
      * \date   23.06.2008
      *
      * \param Pointer to the vector
      * \param Position (zero-based)
      *
      * \return Pointer to vector entry or *null
      */
     P getVectorEntry...
     P                 B
     D                 PI              *
     D   vector                        *   const
     D   pos                         10I 0 const
      *
     D i               S             10I 0
     D header          DS                  likeds(tmpl_header) based(vector)
     D entry           S               *
      /free
       entry = header.elementData + (pos * %size(tmpl_entry));
       return entry;
      /end-free
     P                 E


     /**
      * \brief Check for vector implementation
      *
      * Checks if the pointer points to a vector.
      * The memory of the vector implementation of this service program
      * has an id in the first 20 bytes of the  header.
      *
      * <br><br>
      *
      * If the pointer does not point to a vector an escape message
      * will be sent.
      *
      * \author Mihael Schmidt
      * \date   23.06.2008
      *
      * \param Pointer to the vector
      */
     P isVectorImpl...
     P                 B
     D                 PI
     D   vector                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D valid           S               N
      /free
       monitor;
         if (header.id = VECTOR_ID);
           valid = *on;
         else;
           valid = *off;
         endif;

         on-error *all;
           valid = *off;
       endmon;

       if (not valid);
         sendEscapeMessage(MSG_NO_VECTOR_IMPL);
       endif;
      /end-free
     P                 E


     /**
      * \brief Push element down by x positions
      *
      * Moves the passed element up or down by the x number of entries. This
      * procedure also pushes/pulls every element under it also up or down by
      * the given number of positions.
      *
      * <br><br>
      *
      * A positive <em>downBy</em> value will push the entries down. A negative
      * <em>downBy</em> value will pull the entries up.
      *
      * \author Mihael Schmidt
      * \date   30.07.2008
      *
      * \param Pointer to the vector
      * \param Number of positions to go up or down (default: 1 = down by one position)
      */
     P moveElements...
     P                 B
     D                 PI
     D   vector                        *   const
     D   pos                         10U 0 const
     D   pDownBy                     10I 0 const options(*nopass)
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D downBy          S             10I 0 inz(1)
     D x               S             10I 0
     D elementsToCopy...
     D                 S             10I 0
     D startElement    S               *
     D tmp_mem         S               *
      /free
       if (%parms() = 3);
         downBy = pDownBy;
       endif;

       // check if the destination of the entry is before the head of the vector
       if (pos + downBy < 0);
         sendEscapeMessage(MSG_DEST_OUT_OF_BOUNDS);
       endif;

       // resize the array if necessary
       if (header.elementsAllocated < header.elementCount + downBy);
         increment(vector : header.elementCount + downBy);
       endif;

       startElement = getVectorEntry(vector : pos);

       elementsToCopy = header.elementCount - pos;

       if (startElement <> *null);
         // alloc temp. memory for entries
         cee_getStorage(header.heapId :
                        elementsToCopy * %size(tmpl_entry) :
                        tmp_mem :
                        *omit);

         // copy to temp. memory
         memcpy(tmp_mem :
                startElement :
                elementsToCopy * %size(tmpl_entry));

         // copy to real destination
         memcpy(startElement + (downBy * %size(tmpl_entry)) :  // destination
                tmp_mem :                                      // source
                elementsToCopy * %size(tmpl_entry));           // number of bytes

         // release the temp. memory
         cee_freeStorage(tmp_mem : *omit);
       endif;
      /end-free
     P                 E


     /**
      * \brief Increment array size
      *
      * Increases the size of the vector either by 2 or up to the passed
      * size. The vector size will be decreased if the given size is lower
      * than the current size.
      *
      * <br><br>
      *
      * If the new size cannot be less than the number of elements currently
      * in the vector.
      *
      * \author Mihael Schmidt
      * \date   23.06.2008
      *
      * \param Pointer to the vector
      * \param New size of the vector (default: determined by the header)
      */
     P increment       B
     D                 PI
     D   vector                        *   const
     D   size                        10U 0 const options(*nopass)
      *
     D header          DS                  likeds(tmpl_header) based(vector)
     D newSize         S             10U 0
      /free
       monitor;
         if (%parms() = 1);
           if (header.incrementSize = 0);
             // double the size
             newSize = header.elementsAllocated * 2;
           else;
             newSize = header.incrementSize;
           endif;
         else;
           if (size = 0);
             // double the size
             newSize = header.elementsAllocated * 2;
           elseif (size <= header.elementsAllocated);
             // return if the new size is going to be smaller
             // than the number of elements currently in the vector
             // (same size => do nothing)
             return;
           else;
             newSize = size;
           endif;
         endif;

         on-error *all;
           sendEscapeMessage(MSG_MAX_SIZE_REACHED);
       endmon;

       if (header.elementsAllocated <> newSize);
         cee_reallocateStorage(header.elementData :
                               newSize * %size(tmpl_entry) :
                               *omit);
         header.elementsAllocated = newSize;
       endif;
      /end-free
     P                 E


     /**
      * \brief Send Escape Message
      *
      * Sends an escape message with the specified Id.
      *
      * \author Mihael Schmidt
      * \date   23.06.2008
      *
      * \param Message Id
      */
     P sendEscapeMessage...
     P                 B
     D                 PI
     D   id                          10I 0 const
      *
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D msgdata         S            512A
     D msgkey          S              4A
     D apiError        S            265A
      /free
       if (id = MSG_NO_VECTOR_IMPL);
         msgdata = 'The pointer does not point to a vector data structure.';
       elseif (id = MSG_POSITION_OUT_OF_BOUNDS);
         msgdata = 'The index points outside the vector. No such element in ' +
                   'the vector.';
       elseif (id = MSG_MAX_SIZE_REACHED);
         msgdata = 'The maximum size of the vector has been reached.';
       elseif (id = MSG_DEST_OUT_OF_BOUNDS);
         msgdata = 'The destination for an entry is outside the vector.';
       elseif (id = MSG_RANGE_OUT_OF_BOUNDS);
         msgdata = 'The selected range of elements is out of bounds of the ' +
                   'vector.';
       elseif (id = MSG_INVALID_VALUE_TYPE);
         msgdata = 'The requested type does not correspond to the vector ' +
                   'entry type.';
       endif;

       sendProgramMessage('CPF9898' :
                          'QCPFMSG   *LIBL     ' :
                          %trimr(msgdata) :
                          %len(%trimr(msgdata)) :
                          '*ESCAPE   ' :
                          '*PGMBDY' :
                          0 :
                          msgkey :
                          apiError);
      /end-free
     P                 E
